/*
 *  Name:
 *	dmprel - Dump Microsoft format .rel files.
 *
 *  Usage:
 *	dmprel <infile>
 *
 *  Description:
 *	This program will read a Microsoft format object or library file
 *	and produce an interpreted dump of that file at STDOUT.  The dump
 *	consists of a descriptive header and the item which may consist of
 *	a hexidecimal number and/or an ascii string.
 *
 *  Bugs:
 *	No provision for default file types is provided.
 *
 *  Author:
 *	Frank P. MacLachlan
 *
 *  Edit history:
 *	00 FPM 22-Jul-81	Initial program entry and debug
 */

#include <std.h>

int	nbytes		{ 1 };
char	buf[BUFSIZE]	{ "" };
char	*bp		{ &buf[BUFSIZE-1] };
int 	bitno		{ -1 };
FILE	fn		{ };


/*
 * Main procedure. Open input file and interpret the input bit stream.
 * Special link items require further interpretation.
 */
main(argc, argv)
	int argc;
	char *argv[];
{
	FILE	open();

	if (argc < 2)
		usage();
	if ((fn = open(argv[1], READ, 1)) < 0)
		cant(argv[1]);
	for (;;) {
		if (gnbit() == 0)
			putfmt("\t\tAbsolute byte: %+02hH", getbits(8));
		else
			switch (getbits(2)) {

			case 00:
				getspec();
				break;

			case 01:
				putfmt("\t\tProgram relative word: %+04hH", getwrd());
				break;

			case 02:
				putfmt("\t\tData relative word: %+04hH", getwrd());
				break;

			case 03:
				putfmt("\t\tCOMMON relative word: %+04hH", getwrd());
				break;

			}
		putch('\n');
	}
}		


/*
 * Decode special link items (100xxxx binary)
 */
getspec()
{
	register int len;

	switch (getbits(4)) {


	case 0:
		putfmt("\tEntry symbol: ");
		putb();
		break;

	case 1:
		putfmt("\tSelect COMMON block: ");
		putb();
		break;

	case 2:
		putfmt("Module name: ");
		putb();
		break;

	case 3:
		putfmt("\tRequest library search: ");
		putb();
		break;

	case 4:
		len = getbits(8);
		putfmt("\tExtension link item: %+02hH %+02hH ",
			(len = len ? len : 8), getbits(8));
		while (--len >= 0)
			putfmt("%+02hH%ac", getbits(8), len == 0 ? '\n' : ' ');
		break;

	case 5:
		putfmt("\tDefine COMMON size: ");
		puta();
		putfmt(", ");
		putb();
		break;

	case 6:
		putfmt("\tExternal chain: ");
		puta();
		putfmt(", ");
		putb();
		break;

	case 7:
		putfmt("\tDefine entry point: ");
		puta();
		putfmt(", ");
		putb();
		break;

	case 8:
		putfmt("\tExternal - Offset: ");
		puta();
		break;

	case 9:
		putfmt("\tExternal + Offset: ");
		puta();
		break;

	case 10:
		putfmt("\tDefine Data area size: ");
		puta();
		break;

	case 11:
		putfmt("\tSet loading location counter: ");
		puta();
		break;

	case 12:
		putfmt("\tChain address: ");
		puta();
		break;

	case 13:
		putfmt("\tDefine Program size: ");
		puta();
		break;

	case 14:
		putfmt("\tEnd of module: ");
		puta();
		while (bitno >= 0)
			gnbit();
		putch('\n');
		break;

	case 15:
		putfmt("\n**** End of File ****\n");
		exit();

	}
}


/*
 * Output type 'A' item.
 */
puta()
{
	register int type;
	static char *tp[] {
		"Absolute",
		"Program relative",
		"Data relative",
		"COMMON relative"
	};

	type = getbits(2);
	putfmt("%+04hH (%p)", getwrd(), tp[type]);
}


/*
 * Output type 'B' item.
 */
putb()
{
	register int len;

	putch('"');
	for (len = getbits(3); len > 0; --len)
		putch(getbits(8));
	putch('"');
}


/*
 * Get a word (16 bits) from the input file.
 * First byte read is low byte, second is high byte.
 */
unsigned getwrd()
{

	return (getbits(8) | getbits(8)<<8);
}


/*
 * Get nbits bits from the input file.
 */
unsigned getbits(nbits)
	register int nbits;
{
	register unsigned val;

	for (val = 0; nbits > 0; --nbits)
		val = (val << 1) | gnbit();
	return (val);
}


/*
 * Get the next bit from the input file.
 */
int gnbit()
{
	static char masks[]  { 01, 02, 04, 010, 020, 040, 0100, 0200 };

	if (bitno < 0) {
		if (++bp > &buf[nbytes-1]) {
			bp = &buf;
			if ((nbytes = read(fn, buf, BUFSIZE)) <= 0) {
				errfmt("?DMPREL-F-Unexpected EOF\n");
				exit();
			}
		}
		bitno = 7;
	}
	return ((*bp & masks[bitno--]) != 0);
}	


/*
 * Print diagnostic upon file open error.
 */
cant(fname)
	char *fname;
{

	errfmt("%p: can't open\n", fname);
	exit();
}


/*
 * Tell user how to use this program.
 */
usage()
{
	errfmt("Usage: dmprel <file>\n");
	exit();
}
